-- VECTRIC LUA SCRIPT
-- Copyright 2013 Vectric Ltd.

-- Gadgets are an entirely optional add-in to Vectric's core software products. 
-- They are provided 'as-is', without any express or implied warranty, and you make use of them entirely at your own risk.
-- In no event will the author(s) or Vectric Ltd. be held liable for any damages arising from their use.

-- Permission is granted to anyone to use this software for any purpose, 
-- including commercial applications, and to alter it and redistribute it freely, 
-- subject to the following restrictions:

-- 1. The origin of this software must not be misrepresented; you must not claim that you wrote the original software. If you use this software in a product, an acknowledgment in the product documentation would be appreciated but is not required.
-- 2. Altered source versions must be plainly marked as such, and must not be misrepresented as being the original software.
-- 3. This notice may not be removed or altered from any source distribution.

--require("mobdebug").start()
require "strict"

-- Define global variables
g_version = "1.5"
g_title = "Halftoner"
g_html_file = "halftoner.htm"
g_width = 800
g_height = 760
g_cmUnsetHeight = 9999999
gInches2MM = 25.4
gMM2Inches = 0.03937007874




--[[ GetCuttingDepthFromHeight
|
|   Get the depth at which we must plunge to based on the height of the model at that point
]]
function GetCuttingDepthFromHeight(component, max_height, x, y, max_radius, with_v_bit, tool, options)
  
  if max_height  == 0 then
    return 0
  end
  
  local height = component:GetInterpolated(x,y)
  if height == g_cmUnsetHeight then
    return 0
  end
  
  
  -- scale the height so in [0,1] inverted so high points are cut very deeply?
  local scaled_height = 1 - height / max_height
  local desired_radius = scaled_height*max_radius
  
  -- Get the height we must plunge in to
  local depth = 0
  if (with_v_bit) then
    depth = CalculateDepthForVBit(desired_radius, tool)
  else
    depth = CalculateDepthForBallNose(desired_radius, options.units_corrected_tool_rad)
  end

  return depth
end





 --[[ CalculateDepthForBallNose
 |
 | Computes the depth of the cut needed to achieve the given radius.
 ]]
 function CalculateDepthForBallNose(radius, tool_radius)
  local chord_radius = radius
  local s = math.sqrt(math.pow(tool_radius, 2) - math.pow(chord_radius, 2))
  local depth = tool_radius - s
  return depth
end




--[[ CalculateDepthForVBit
|
| Compute the depth of the cut needed to achieve the given radius.
| I think the formula is d = r/tan(t/2)/
]]
function CalculateDepthForVBit(radius, tool)
  local angle = tool.VBit_Angle
  local depth = radius / (math.tan(math.rad(angle/2)))
  return depth
end






--[[ main
|
| Enter the program here
]]
function main(script_path)
  
  -- Set up options --
  if not IsAspire() then
  	DisplayMessageBox("This gadget requires Aspire")
  	return true
  end

  if (GetAppVersion() < 4.5) then
  	DisplayMessageBox("This gadget requires a version of Aspire greater than 4.5")
  	return true
  end
  
  -- Check for the existence of a job
local job = VectricJob()
if not job.Exists then
  	MessageBox("No job created, please start a new job and re-run the gadget")
  return false
end
  
  
  local options = {}
  options.toolpath_name = "Halftone toolpath"
  options.start_depth = 0
  options.retract_z = 1
  options.max_height = 0
  options.max_radius = 16
  options.units_corrected_tool_rad = 0
  options.safe_z = 1
  options.tool = nil
  options.angle  = 60
  options.comp_angle = 60
  options.units_string = ""
  
  
  
  LoadDefaults(options)
  
  -- Get a bounding box for our selections.
  local vectric_job = VectricJob()
  local selection = vectric_job.Selection
  if vectric_job.InMM then
    options.units_string  = "mm"
  else
    options.units_string = "inches"
  end
  
  local component_manager = vectric_job.ComponentManager
  local selected_object_ids = component_manager:GetSelectedObjectIds(true) -- We want a list of all selected nodes
  
  
  
 --[[ if selected_object_ids.IsEmpty then
    MessageBox("nothing")
    
  else
    MessageBox("nothing2")
  end--]]
  
  -- Create a local relief
  local relief = component_manager:CreateReliefFromCompositeModel()
  local _, min, max = relief:GetMinMaxZ()
  options.max_height = max
  
  if relief.IsTransparent then
    MessageBox("No component detected, create component and run Gadget again")
    return false
    end
  --MessageBox(tostring(min) .. " " ..tostring(max))
 
  -- Display the dialog
  local displayed_dialog = -1
  while displayed_dialog == -1 do
    displayed_dialog = DisplayDialog(script_path, options)
  end
  
  if not displayed_dialog then
    return false
  end
  
  -- Get the options from the dialog
  local tool = options.tool
  local use_selection_order = options.use_selection_order
  
  -- prepare the tool
  local tool_type = tool.ToolType -- should be Tool.VBIT or Tool.BALL_NOSE
  if not (tool_type == Tool.VBIT or tool_type == Tool.BALL_NOSE) then
    DisplayMessageBox("Selected tool must be Vbit or ball-nose")
    return false
  end
  
 options.units_corrected_tool_rad = (tool.ToolDia / 2)

 if tool.InMM and (not vectric_job.InMM) then
 	options.units_corrected_tool_rad = (tool.ToolDia /2) * gMM2Inches
 end

 if vectric_job.InMM and (not tool.InMM) then
 	options.units_corrected_tool_rad = (tool.ToolDia / 2) * gInches2MM
 end

 if options.max_radius > options.units_corrected_tool_rad then
   MessageBox("You have chosen a maximum hole radius which is which is more than the radius of the tool you are using. Please choose a smaller radius.")
   displayed_dialog = -1
  end
  
  -- Create the depth and point array
  local depth_array = {}
  local point_array = {}
   

  -- Get the bounds off the whole job
  local bounds = vectric_job:GetBounds()
  local start_point = bounds.BLC
  -- first go along in x
  local shift_in_x  = 2*options.max_radius * Vector2D(math.cos(math.rad(options.comp_angle)), -math.sin(math.rad(options.comp_angle)))
  while start_point.x  < bounds.MaxX do
    start_point = CreateHalftonePointsLine(relief, point_array, depth_array, options, start_point, bounds)
    if start_point == nil then
      break;
    end
    start_point = start_point + shift_in_x
  end

  -- now go up in y
  local shift_in_y = 2*options.max_radius * Vector2D(- math.cos(math.rad(options.comp_angle)), math.sin(math.rad(options.comp_angle)));
  start_point = bounds.BLC
  while start_point.y < bounds.MaxY do
    start_point = CreateHalftonePointsLine(relief, point_array, depth_array, options, start_point, bounds)
    if start_point == nil then
      break;
    end
    start_point = start_point + shift_in_y
  end
  
  local pos_data = ToolpathPosData()
  options.safe_z = pos_data.SafeZ;
  pos_data:EnsureHomeZIsSafe()
  
  
  -- Create a contour for these points
  local contour_grp = ContourGroup(true)
  local ctr = Contour(0.0)
  ctr = CreateToolpathContour(depth_array, point_array, options)
  if not ctr then
    DisplayMessageBox("Unable to create toolpath contour.")
    return false
  end
  
  contour_grp:AddTail(ctr)
 
  -- Set external toolpath options.
  local toolpath_options = ExternalToolpathOptions()
  toolpath_options.StartDepth = 0
  toolpath_options.CreatePreview = true
  
  -- create toolpath from the created contour
  local toolpath = ExternalToolpath(
                                      options.toolpath_name,
                                      tool,
                                      pos_data,
                                      toolpath_options,
                                      contour_grp
                                    )

                                    
  if not toolpath then
    DisplayMessageBox("Unable to make contour")
    return false
  end
  
  -- Add the toolpath to the toolpath manager
  AddExternalToolpaths({toolpath})
  SaveDefaults(options)
  return true
end





--[[ CreateHalftonePointsLine
    
    Create a single line of drilling points.
    Store in point array and depth array.
--]]
function CreateHalftonePointsLine(model, point_array, depth_array, options, start_pt, bounds)
  -- we store the new line in a new array and decide to attach those points in the order
  -- we specify

  local this_point_array = {}
  local this_depth_array = {}

  local shift_vector = 2*options.max_radius * Vector2D(math.cos(math.rad(options.angle)), math.sin(math.rad(options.angle)))
  local timeout = 10
  while (not bounds:IsInside(start_pt,0)) and (timeout > 0)  do
    start_pt = start_pt + shift_vector
    timeout = timeout -1
  end
  
  if timeout == 0 then
    return nil
  end
  
  local next_pt = start_pt
  while bounds:IsInside(next_pt, 0.0) do
    next_pt = next_pt + shift_vector
    
    local height =  GetCuttingDepthFromHeight(model, 
                      options.max_height, 
                      next_pt.x, 
                      next_pt.y, 
                      options.max_radius, 
                      options.with_v_bit,
                      options.tool,
                      options
                    )
    -- Append the point and the depth to the appropriate arrays
    this_depth_array[#this_depth_array + 1] = height
    this_point_array[#this_point_array + 1] = next_pt
  end

  -- check whether the last point of the point array is closer to 
  -- the start or end of the array. Join in this way
  local first_added_point = this_point_array[1]
  local last_added_point = this_point_array[#this_point_array]
  local last_point = point_array[#point_array]
  local reverse_array = false;
  if (last_point ~= nil) then
    local first_dist = DistBetweenPtsSq(first_added_point, last_point)
    local last_dist = DistBetweenPtsSq(last_added_point, last_point)
    if (first_dist > last_dist ) then
      reverse_array = true
    end
  end

  if not reverse_array then
    for i = 1, #this_point_array do
      point_array[#point_array + 1] = this_point_array[i]
      depth_array[#depth_array + 1] = this_depth_array[i]    
    end
  else
    for i = #this_point_array, 1, -1 do
      point_array[#point_array + 1] = this_point_array[i]
      depth_array[#depth_array + 1] = this_depth_array[i]    
    end
  end
  last_point = point_array[#point_array]
  return start_pt
 
end




--[[ DistBetweenPtsSq
    Return the square distance between two points
--]]
function DistBetweenPtsSq(a, b)
  return (a.x - b.x) * (a.x - b.x) + (a.y - b.y)*(a.y - b.y)
end





--[[ CreateToolpathContour
     Given two arrays of depths and points then we create a toolpath.
--]]
function CreateToolpathContour(depth_array, point_array, options)
  local ctr = Contour(0.0);
 
  -- Do start data
  ctr:AppendPoint(0,0, options.safe_z);
  
  -- iterate over the points
  for i, _ in pairs(depth_array) do
    if depth_array[i] == 0 then
    else
      local start_pt = Point3D(point_array[i], options.retract_z)
      local plunge_pt = Point3D(point_array[i], -depth_array[i])
      ctr:LineTo(start_pt);
      ctr:LineTo(plunge_pt);
      ctr:LineTo(start_pt);
    end
  end
  
  local safe_end = Point3D(0,0, options.safe_z)
  ctr:LineTo(safe_end) 
  return ctr
end






--[[ -------- DisplayDialog -------------------
|     
|     Display the dialog
]]
function DisplayDialog(script_path, options)
  local html_path = "file:" .. script_path .. "\\" .. g_html_file
  
  local dialog = HTML_Dialog(false, html_path, g_width, g_height, "Halftone VCarve Gadget")
  
  -- add fields to add version and title to header
  dialog:AddLabelField("GadgetTitle", g_title)
  dialog:AddLabelField("GadgtVersion", g_version)
  
  -- Add toolpicker
  local tooldb = ToolDatabase()
  dialog:AddLabelField("ToolNameLabel", "")
  dialog:AddToolPicker("ToolChooseButton", "ToolNameLabel", "")
  dialog:AddToolPickerValidToolType("ToolChooseButton", Tool.VBIT)
  dialog:AddToolPickerValidToolType("ToolChooseButton", Tool.BALL_NOSE)
  
  -- Add double fields 
  dialog:AddDoubleField("MaxRadius", options.max_radius)
  dialog:AddDoubleField("ScreenAngle", options.angle)
  dialog:AddDoubleField("RetractZ", options.retract_z)

  -- Add a field asking the user for the tool path name?
  dialog:AddTextField("ToolpathName", options.toolpath_name)
  
  dialog:AddLabelField("Units1", options.units_string);
  -- Display the dialog  
  local success = dialog:ShowDialog()
  
  if not success then
    return false
  end
  
  -- Get from the dialogs form elements and insert into options table
  options.tool = dialog:GetTool("ToolChooseButton")
  if options.tool == nil then
    DisplayMessageBox("No tool selected.")
    return -1
  end
  
  
  options.toolpath_name = dialog:GetTextField("ToolpathName")
  options.max_radius = dialog:GetDoubleField("MaxRadius")
  options.angle = dialog:GetDoubleField("ScreenAngle")
  options.retract_z = dialog:GetDoubleField("RetractZ")
  options.with_v_bit = options.tool.ToolType == Tool.VBIT
  return true
  
end




--[[ -------- AddExternalToolpaths---------------------------------  
|
|  Add toolpaths to the toolpath manager.
|  
|
]]
function AddExternalToolpaths(toolpaths)
  -- check whether each toolpath can be added to the toolpath manager.
  local toolpath_manager = ToolpathManager()
  for _,toolp in ipairs(toolpaths) do
    local success = toolpath_manager:AddExternalToolpath(toolp)
    if not success then
      DisplayMessageBox("Unable to add the toolpaths to the toolpath manager")
      return false
    end
  end
  return true
end



--[[ -------- SaveDefaults---------------------------------  
|
|  Save Default options to the registry
|  
|
]]
function SaveDefaults(options)
  local registry = Registry("HalftoneVCarve")
  registry:SetDouble("MaxRadius", options.max_radius)
  registry:SetDouble("RetractZ", options.retract_z)
  registry:SetDouble("Angle", options.angle)
end

--[[ -------- LoadDefaults---------------------------------  
|
|  Load Default options from the registry
|     
|
]]
function LoadDefaults(options)
  local registry = Registry("HalftoneVCarve")
  local max_radius = registry:GetDouble("MaxRadius", 0)
  local retract_z = registry:GetDouble("RetractZ", 1)
  local angle = registry:GetDouble("Angle", 60)

  options.max_radius = max_radius
  options.retract_z = retract_z
  options.angle = angle
end
  
